#ifndef TX_COMMON_H
#define TX_COMMON_H

#include <cstddef>
#include <stdio.h>
#include <iostream>
#include <algorithm> // std::min
#include <cuda.h>
#include <cuComplex.h>
#ifdef CUSP_USE_TEXTURE_MEMORY
#include <cusp/detail/device/texture.h>
#endif
#include <cusp/detail/device/arch.h>
#include <cusp/detail/device/common.h>
#include <cusp/detail/device/utils.h>
#include <cusp/version.h>

#include <cusparse_v2.h>

static inline char * cusparseGetErrorString(cusparseStatus_t stat) {
  switch (stat) {
  case CUSPARSE_STATUS_NOT_INITIALIZED:
    return const_cast<char *>(" CuSPARSE not initialized");
  case CUSPARSE_STATUS_ALLOC_FAILED:
    return const_cast<char *>(" CuSPARSE allocation failed");
  case CUSPARSE_STATUS_INVALID_VALUE:
    return const_cast<char *>(" CuSPARSE invalid input argument");
  case CUSPARSE_STATUS_ARCH_MISMATCH:
    return const_cast<char *>(" CuSPARSE architecture mismatch");
  case CUSPARSE_STATUS_MAPPING_ERROR:
    return const_cast<char *>(" CuSPARSE mapping error");
  case CUSPARSE_STATUS_EXECUTION_FAILED:
    return const_cast<char *>(" CuSPARSE execution failed");
  case CUSPARSE_STATUS_INTERNAL_ERROR:
    return const_cast<char *>(" CuSPARSE internal error");
  case CUSPARSE_STATUS_MATRIX_TYPE_NOT_SUPPORTED:
    return const_cast<char *>(" CuSPARSE matrix type not supported");
  default:
    return const_cast<char *>(" CuSPARSE success");
  }  
}

#define CUSPARSE_ERR_CHK(stat, tx_function, cusparse_function) if (stat != CUSPARSE_STATUS_SUCCESS) {std::cout<<"Failed in "<<tx_function<<" ... "<<cusparse_function<<" : "<<cusparseGetErrorString(stat)<<std::endl; return stat; }
#define CUDA_ERR_CHK(err, tx_function, cuda_function) if (err != cudaSuccess) {std::cout<<"Failed in "<<tx_function<<" ... "<<cuda_function<<" ... cuda error "<<cudaGetErrorString(err) <<std::endl; return err; }

#define CUSPARSE_ERR_CHK_NORETURN(stat, tx_function, cusparse_function) if (stat != CUSPARSE_STATUS_SUCCESS) {std::cout<<"Failed in "<<tx_function<<" ... "<<cusparse_function<<" : " <<cusparseGetErrorString(stat)<<std::endl;}
#define CUDA_ERR_CHK_NORETURN(err, tx_function, cuda_function) if (err != cudaSuccess) {std::cout<<"Failed in "<<tx_function<<" ... "<<cuda_function<<" ... cuda error "<<cudaGetErrorString(err) <<std::endl;}

#if THRUST_VERSION >= 100500
#include <thrust/detail/backend/cuda/arch.h>
using namespace thrust::detail::backend::cuda::arch;
#else
#include <thrust/detail/device/cuda/arch.h>
using namespace thrust::detail::device::cuda::arch;
#endif

#define CUSPINTARRAYGPU32 cusp::array1d<int,cusp::device_memory>
#define CUSPINTARRAYCPU32 cusp::array1d<int,cusp::host_memory>

#define DIVIDE_INTO(x,y) ((x + y - 1)/y)

#define MAT_OPERATION cusparseOperation_t 
#define NON_TRANSPOSE CUSPARSE_OPERATION_NON_TRANSPOSE
#define TRANSPOSE CUSPARSE_OPERATION_TRANSPOSE
#define CONJUGATE_TRANSPOSE CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE

template<typename index_type, typename value_type>
  struct cscMatrix {
    index_type num_rows;
    index_type num_cols;
    index_type num_entries;
    CUSPARRAY * values;
    CUSPINTARRAYGPU32 * row_indices;
    CUSPINTARRAYGPU32 * column_offsets;
  };

template<typename index_type, typename value_type>
  struct csrMatrix {
    index_type num_rows;
    index_type num_cols;
    index_type num_entries;
    CUSPARRAY * values;
    CUSPINTARRAYGPU32 * row_offsets;
    CUSPINTARRAYGPU32 * column_indices;
  };


template<typename value_type>
class Equals {
 public:
  __device__ value_type operator() (value_type a, value_type b) {
    return b;
  }
};

template<typename value_type>
class PlusEquals {
 public:
  __device__ value_type operator() (value_type a, value_type b) {
    return a+b;
  }
};


struct ThrustPlusEquals
{
  template <typename Tuple>
  __host__ __device__
  void operator()(Tuple t)
  {
    thrust::get<1>(t) = thrust::get<1>(t) + thrust::get<0>(t);
  }
};


template <typename value_type>
__host__ __device__ static __inline__ value_type _fma_(const value_type a, value_type b, value_type c){

#if defined(PETSC_USE_COMPLEX)
#if defined(PETSC_USE_REAL_SINGLE)
  return cuCfmaf(a, b, c);
#elif defined(PETSC_USE_REAL_DOUBLE)
  return cuCfma(a, b, c);
#endif

#else
  return a*b + c;
#endif
};

template <typename value_type>
__host__ __device__ static __inline__ value_type _zero_(){

#if defined(PETSC_USE_COMPLEX)

#if defined(PETSC_USE_REAL_SINGLE)
  return make_cuFloatComplex(0.f, 0.f);
#elif defined(PETSC_USE_REAL_DOUBLE)
  return make_cuDoubleComplex(0., 0.);
#endif

#else

#if defined(PETSC_USE_REAL_SINGLE)
  return 0.f;
#elif defined(PETSC_USE_REAL_DOUBLE)
  return 0.;
#endif

#endif
};


template<typename index_type, typename value_type>
  void transpose(const int m, const int nnz,
		 const index_type * row_offsets, const index_type * column_indices, const value_type * values,
		 index_type * row_offsets_transpose, index_type * column_indices_transpose, value_type * values_transpose) {

  // Need an offset into the column
  int * currentPosInColumn = (int*) malloc(m*sizeof(int));
  memset(currentPosInColumn, 0, m*sizeof(int));
  memset(row_offsets_transpose, 0, (m+1)*sizeof(PetscInt));

  for (int j=0; j<m; j++) {
    for (int i=row_offsets[j]; i<row_offsets[j+1]; i++) {
      row_offsets_transpose[column_indices[i]+1]++;
    }
  }

  // cumulative sums
  for (int j=0; j<m; j++) {
    row_offsets_transpose[j+1]+=row_offsets_transpose[j];
  }

  for (int j=0; j<m; j++) {
    for (int i=row_offsets[j]; i<row_offsets[j+1]; i++) {
      int col_index = column_indices[i];
      int row_off = row_offsets_transpose[col_index];
      column_indices_transpose[row_off + currentPosInColumn[col_index]]=j;
      values_transpose[row_off + currentPosInColumn[col_index]]=values[i];
      currentPosInColumn[col_index]++;
    }
  }
  // release memory
  free(currentPosInColumn);
}

template<typename index_type, typename value_type>
  void convertSymmToFullMatrix(const index_type m, const index_type nnz,
			       const index_type * row_offsets, const index_type * column_indices,
			       const value_type * values, index_type * row_offsets_full, 
			       index_type * column_indices_full, value_type * values_full) {

  index_type * row_offsets_transpose;
  index_type * column_indices_transpose;
  value_type * values_transpose;
  index_type * num_per_row_full;

  row_offsets_transpose = (index_type *) malloc((m+1)*sizeof(index_type));
  memset(row_offsets_transpose, 0, (m+1)*sizeof(index_type));

  num_per_row_full = (index_type *) malloc(m*sizeof(index_type));
  memset(num_per_row_full, 0, m*sizeof(index_type));
  
  column_indices_transpose = (index_type *) malloc(nnz*sizeof(index_type));
  memset(column_indices_transpose, 0, nnz*sizeof(index_type));
  
  values_transpose = (value_type *) malloc(nnz*sizeof(value_type));
  memset(values_transpose, 0, nnz*sizeof(value_type));    

  // first, transpose the matrix.
  transpose<index_type, value_type>(m, nnz, row_offsets, column_indices, values,
				    row_offsets_transpose, column_indices_transpose, values_transpose);

  // now calculate the number per row ... subtract one in the transposed case as to not 
  // include the diagonal twice.
  for (int j=0; j<m; j++) {
    num_per_row_full[j] = (row_offsets[j+1]-row_offsets[j]) + (row_offsets_transpose[j+1]-row_offsets_transpose[j]-1);
  }
  // compute the row_offsets of the full matrix
  for (int j=0; j<m; j++) {
    row_offsets_full[j+1] = row_offsets_full[j] + num_per_row_full[j];
  }

  // now combine the original and the transpose matrices into the full matrix
  int k=0;
  for (int j=0; j<m; j++) {
    // in this loop, we make sure to not include the diagonal part of the matrix.
    for (int i=row_offsets_transpose[j]; i<row_offsets_transpose[j+1]-1; i++) {
      column_indices_full[k] = column_indices_transpose[i];
      values_full[k] = values_transpose[i];
      k++;
    }

    for (int i=row_offsets[j]; i<row_offsets[j+1]; i++) {
      column_indices_full[k] = column_indices[i];
      values_full[k] = values[i];
      k++;
    }
  }

  free(row_offsets_transpose);
  free(column_indices_transpose);
  free(values_transpose);
  free(num_per_row_full);
}


template<typename index_type, typename value_type>
  void convertHermToFullMatrix(const index_type m, const index_type nnz,
			       const index_type * row_offsets, const index_type * column_indices,
			       const value_type * values, index_type * row_offsets_full, 
			       index_type * column_indices_full, value_type * values_full) {

}

#endif // TX_COMMON_H
